package app.simple.inure.text;

/*
 * THIS CLASS IS PROVIDED TO THE PUBLIC DOMAIN FOR FREE WITHOUT ANY
 * RESTRICTIONS OR ANY WARRANTY.
 */

import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.text.Editable;
import android.text.Selection;
import android.text.SpannableStringBuilder;
import android.text.TextWatcher;
import android.text.style.UnderlineSpan;
import android.widget.TextView;

import java.util.LinkedList;

/**
 * A generic undo/redo implementation for TextViews.
 */
public class TextViewUndoRedo {
    
    /**
     * Is undo/redo being performed? This member signals if an undo/redo
     * operation is currently being performed. Changes in the text during
     * undo/redo are not recorded because it would mess up the undo history.
     */
    private boolean isUndoOrRedo = false;
    
    /**
     * The edit history.
     */
    private final EditHistory editHistory;
    
    /**
     * The change listener.
     */
    private final EditTextChangeListener editTextChangeListener;
    
    /**
     * The edit text.
     */
    private final TextView textView;
    
    // =================================================================== //
    
    /**
     * Create a new TextViewUndoRedo and attach it to the specified TextView.
     *
     * @param textView The text view for which the undo/redo is implemented.
     */
    public TextViewUndoRedo(TextView textView) {
        this.textView = textView;
        editHistory = new EditHistory();
        editTextChangeListener = new EditTextChangeListener();
        this.textView.addTextChangedListener(editTextChangeListener);
    }
    
    // =================================================================== //
    
    /**
     * Disconnect this undo/redo from the text view.
     */
    public void disconnect() {
        textView.removeTextChangedListener(editTextChangeListener);
    }
    
    /**
     * Set the maximum history size. If size is negative, then history size is
     * only limited by the device memory.
     */
    public void setMaxHistorySize(int maxHistorySize) {
        editHistory.setMaxHistorySize(maxHistorySize);
    }
    
    /**
     * Clear history.
     */
    public void clearHistory() {
        editHistory.clear();
    }
    
    /**
     * Can undo be performed?
     */
    public boolean getCanUndo() {
        return (editHistory.mmPosition > 0);
    }
    
    /**
     * Perform undo.
     */
    public void undo() {
        EditItem edit = editHistory.getPrevious();
        if (edit == null) {
            return;
        }
    
        Editable text = textView.getEditableText();
        int start = edit.mmStart;
        int end = start + (edit.mmAfter != null ? edit.mmAfter.length() : 0);
    
        isUndoOrRedo = true;
        text.replace(start, end, edit.mmBefore);
        isUndoOrRedo = false;
    
        /*
         * This will get rid of underlines inserted when editor tries to come
         * up with a suggestion.
         */
        for (UnderlineSpan underlineSpan : text.getSpans(0, text.length(), UnderlineSpan.class)) {
            text.removeSpan(underlineSpan);
        }
    
        Selection.setSelection(text, edit.mmBefore == null ? start : (start + edit.mmBefore.length()));
    }
    
    /**
     * Can redo be performed?
     */
    public boolean getCanRedo() {
        return (editHistory.mmPosition < editHistory.mmHistory.size());
    }
    
    /**
     * Perform redo.
     */
    public void redo() {
        EditItem edit = editHistory.getNext();
        if (edit == null) {
            return;
        }
    
        Editable text = textView.getEditableText();
        int start = edit.mmStart;
        int end = start + (edit.mmBefore != null ? edit.mmBefore.length() : 0);
    
        isUndoOrRedo = true;
        text.replace(start, end, edit.mmAfter);
        isUndoOrRedo = false;
    
        /*
         * This will get rid of underlines inserted when editor tries to come
         * up with a suggestion.
         */
        for (UnderlineSpan span : text.getSpans(0, text.length(), UnderlineSpan.class)) {
            text.removeSpan(span);
        }
    
        Selection.setSelection(text, edit.mmAfter == null ? start : (start + edit.mmAfter.length()));
    }
    
    /**
     * Store preferences.
     */
    public void storePersistentState(Editor editor, String prefix) {
        // Store hash code of text in the editor so that we can check if the
        // editor contents has changed.
        editor.putString(prefix + ".hash",
                String.valueOf(textView.getText().toString().hashCode()));
        editor.putInt(prefix + ".maxSize", editHistory.mmMaxHistorySize);
        editor.putInt(prefix + ".position", editHistory.mmPosition);
        editor.putInt(prefix + ".size", editHistory.mmHistory.size());
    
        int i = 0;
        for (EditItem ei : editHistory.mmHistory) {
            String pre = prefix + "." + i;
        
            editor.putInt(pre + ".start", ei.mmStart);
            editor.putString(pre + ".before", ei.mmBefore.toString());
            editor.putString(pre + ".after", ei.mmAfter.toString());
        
            i++;
        }
    }
    
    /**
     * Restore preferences.
     *
     * @param prefix The preference key prefix used when state was stored.
     * @return did restore succeed? If this is false, the undo history will be
     * empty.
     */
    public boolean restorePersistentState(SharedPreferences sp, String prefix)
            throws IllegalStateException {
    
        boolean ok = doRestorePersistentState(sp, prefix);
        if (!ok) {
            editHistory.clear();
        }
    
        return ok;
    }
    
    private boolean doRestorePersistentState(SharedPreferences sharedPreferences, String prefix) {
        String hash = sharedPreferences.getString(prefix + ".hash", null);
        if (hash == null) {
            // No state to be restored.
            return true;
        }
        
        if (Integer.parseInt(hash) != textView.getText().toString().hashCode()) {
            return false;
        }
        
        editHistory.clear();
        editHistory.mmMaxHistorySize = sharedPreferences.getInt(prefix + ".maxSize", -1);
        
        int count = sharedPreferences.getInt(prefix + ".size", -1);
        if (count == -1) {
            return false;
        }
        
        for (int i = 0; i < count; i++) {
            String pre = prefix + "." + i;
            
            int start = sharedPreferences.getInt(pre + ".start", -1);
            SpannableStringBuilder before = new SpannableStringBuilder(sharedPreferences.getString(pre + ".before", null));
            SpannableStringBuilder after = new SpannableStringBuilder(sharedPreferences.getString(pre + ".after", null));
            
            if (start == -1 || before == null || after == null) {
                return false;
            }
            
            editHistory.add(new EditItem(start, before, after));
        }
        
        editHistory.mmPosition = sharedPreferences.getInt(prefix + ".position", -1);
        return editHistory.mmPosition != -1;
    }
    
    public void addHistory(int start, Editable beforeChange, Editable afterChange) {
        editHistory.add(new EditItem(start, beforeChange, afterChange));
    }
    
    // =================================================================== //
    
    /**
     * Keeps track of all the edit history of a text.
     */
    private static final class EditHistory {
        
        /**
         * The position from which an EditItem will be retrieved when getNext()
         * is called. If getPrevious() has not been called, this has the same
         * value as mmHistory.size().
         */
        private int mmPosition = 0;
        
        /**
         * Maximum undo history size.
         */
        private int mmMaxHistorySize = -1;
        
        /**
         * The list of edits in chronological order.
         */
        private final LinkedList <EditItem> mmHistory = new LinkedList <>();
        
        /**
         * Clear history.
         */
        private void clear() {
            mmPosition = 0;
            mmHistory.clear();
        }
        
        /**
         * Adds a new edit operation to the history at the current position. If
         * executed after a call to getPrevious() removes all the future history
         * (elements with positions >= current history position).
         */
        private void add(EditItem item) {
            while (mmHistory.size() > mmPosition) {
                mmHistory.removeLast();
            }
            mmHistory.add(item);
            mmPosition++;
            
            if (mmMaxHistorySize >= 0) {
                trimHistory();
            }
        }
        
        /**
         * Set the maximum history size. If size is negative, then history size
         * is only limited by the device memory.
         */
        private void setMaxHistorySize(int maxHistorySize) {
            mmMaxHistorySize = maxHistorySize;
            if (mmMaxHistorySize >= 0) {
                trimHistory();
            }
        }
        
        /**
         * Trim history when it exceeds max history size.
         */
        private void trimHistory() {
            while (mmHistory.size() > mmMaxHistorySize) {
                mmHistory.removeFirst();
                mmPosition--;
            }
            
            if (mmPosition < 0) {
                mmPosition = 0;
            }
        }
        
        /**
         * Traverses the history backward by one position, returns and item at
         * that position.
         */
        private EditItem getPrevious() {
            if (mmPosition == 0) {
                return null;
            }
            mmPosition--;
            return mmHistory.get(mmPosition);
        }
        
        /**
         * Traverses the history forward by one position, returns and item at
         * that position.
         */
        private EditItem getNext() {
            if (mmPosition >= mmHistory.size()) {
                return null;
            }
            
            EditItem item = mmHistory.get(mmPosition);
            mmPosition++;
            return item;
        }
    }
    
    /**
     * Represents the changes performed by a single edit operation.
     */
    private static final class EditItem {
        private final int mmStart;
        private final Editable mmBefore;
        private final Editable mmAfter;
        
        /**
         * Constructs EditItem of a modification that was applied at position
         * start and replaced CharSequence before with CharSequence after.
         */
        public EditItem(int start, Editable before, Editable after) {
            mmStart = start;
            mmBefore = before;
            mmAfter = after;
        }
    }
    
    /**
     * Class that listens to changes in the text.
     */
    private final class EditTextChangeListener implements TextWatcher {
    
        /**
         * The text that will be removed by the change event.
         */
        private Editable beforeChange;
    
        /**
         * The text that was inserted by the change event.
         */
        private Editable afterChange;
    
        public void beforeTextChanged(CharSequence s, int start, int count, int after) {
            if (isUndoOrRedo) {
                return;
            }
        
            beforeChange = (Editable) s.subSequence(start, start + count);
        }
    
        public void onTextChanged(CharSequence charSequence, int start, int before, int count) {
            if (isUndoOrRedo) {
                return;
            }
        
            afterChange = (Editable) charSequence.subSequence(start, start + count);
            editHistory.add(new EditItem(start, beforeChange, afterChange));
        }
    
        public void afterTextChanged(Editable s) {
            /*
             * No op
             */
        }
    }
}